TREE_IDENTIFIER = "tree_id"


class GRAPH:
    class ATTRIBUTES:
        CREATED_AT = "created_at"

    class LABELS:
        MENTIONS = "mentions"
        PARENT_OF = "parent_of"
        AUTHOR_OF = "author_of"
        ANSWERED_BY = "answered_by"

    class SUBSETS:
        TWEETS = "tweets"
        AUTHORS = "authors"


class TABLE:
    class COLUMNS:
        PARENT_ID = "parent_id"
        CREATED_AT = "created_at"
        AUTHOR_ID = "author_id"
        TEXT = "text"
        POST_ID = "post_id"
        TREE_ID = "tree_id"



from typing import List

from delab_trees.constants import TABLE


class DelabPost:
    def __init__(self, post_id, parent_id, text: str, tree_id, author_id, created_at):
        self.post_id = post_id
        self.parent_id = parent_id
        self.text = text
        self.tree_id = tree_id
        self.author_id = author_id
        self.created_at = created_at


class DelabPosts:

    @staticmethod
    def from_pandas(df) -> List[DelabPost]:
        result = []
        for i in df.index:
            row = df.loc[i]
            post_id = row[TABLE.COLUMNS.POST_ID]
            parent_id = row[TABLE.COLUMNS.PARENT_ID]
            text = row[TABLE.COLUMNS.TEXT]
            tree_id = row[TABLE.COLUMNS.TREE_ID]
            author_id = row[TABLE.COLUMNS.AUTHOR_ID]
            created_at = row[TABLE.COLUMNS.CREATED_AT]
            post = DelabPost(post_id, parent_id, text, tree_id, author_id, created_at)
            result.append(post)
        return result



import logging

from delab_trees.delab_tree import TABLE
import xml.etree.ElementTree as ET

logger = logging.getLogger(__name__)


class TreeNode:
    def __init__(self, data, post_id: int, parent=None, parent_type="replied_to"):
        self.post_id = post_id
        self.data = data
        self.children = []
        self.max_path_length = 0
        self.parent = parent
        self.parent_type = parent_type

    def find_parent_of(self, node):
        if type(self.post_id) is not type(node.parent.post_id):
            self.post_id = int(self.post_id)
            node.parent.post_id = int(node.parent.post_id)
        assert type(self.post_id) is type(node.parent.post_id)
        if node.parent.post_id == self.post_id:
            self.children.append(node)
            return True
        else:
            for child in self.children:
                result = child.find_parent_of(node)
                if result:
                    return result
        return False

    def print_tree(self, level):
        """level 0 is the root node, then incremented for subsequent generations"""
        print(f'{level * "_"}{level}: {self.data}')
        level += 1
        for child in self.children:
            child.print_tree(level)

    def to_string(self, level=0):
        result = ""
        if level == 0:
            result += "Conversation: " + str(self.data[TABLE.COLUMNS.TREE_ID]) + "\n\n"
        result += (level * "\t") + self.data_to_string(level)
        for child in self.children:
            result += child.to_string(level + 1)
        return result

    def data_to_string(self, level):
        text = self.data[TABLE.COLUMNS.TEXT].split(".")
        tabbed_text = []
        for sentence in text:
            sentence = sentence.replace('\n', ' ').replace('\r', '')
            tabbed_sentence = ""
            if len(sentence) > 125:
                tabbed_sentence += sentence[0:125]
                tabbed_sentence += "\n" + (level * "\t")
                tabbed_sentence += sentence[125:]
            else:
                tabbed_sentence = sentence
            tabbed_text.append(tabbed_sentence)
        separator = ".\n" + (level * "\t")
        tabbed_text = "\n" + (level * "\t") + separator.join(tabbed_text)
        return str(self.data.get("tw_author__name", "namenotgiven")) + "/" + str(
            self.data.get("tw_author__location", "locationnotgiven")) + "/" + str(
            self.data[TABLE.COLUMNS.AUTHOR_ID]) + ":" + tabbed_text + "\n\n"

    def to_norm_xml(self, level=0):
        discourse_elem = ET.Element('discourse')
        platform_elem = ET.SubElement(discourse_elem, 'platform')
        platform_elem.text = self.data['platform']
        speech_acts_elem = ET.SubElement(discourse_elem, 'speech-acts')
        self.tweet_to_speech_act_xml(speech_acts_elem)
        return ET.tostring(discourse_elem, encoding='utf-8')

    def tweet_to_speech_act_xml(self, parent_elem):
        speech_act_elem = ET.SubElement(parent_elem, 'speech-act')
        speech_act_id_elem = ET.SubElement(speech_act_elem, 'speech-act-id')
        speech_act_id_elem.text = str(self.post_id)
        author_elem = ET.SubElement(speech_act_elem, 'author')
        author_id_elem = ET.SubElement(author_elem, 'author-id')
        author_id_elem.text = str(self.data["author_id"])
        author_name_elem = ET.SubElement(author_elem, 'author-name')
        author_name_elem.text = self.data["tw_author__name"]
        text_elem = ET.SubElement(speech_act_elem, 'text')
        text_elem.text = self.data["text"]
        in_response_elem = ET.SubElement(speech_act_elem, 'in-reply-to')
        in_response_elem.text = str(self.data["tn_parent"])
        for child in self.children:
            child.tweet_to_speech_act_xml(speech_act_elem)

    def flat_size(self):
        children_size = 0
        for child in self.children:
            children_size += child.flat_size()
        return 1 + children_size

    def compute_max_path_length(self, level=0):
        # print(level)
        if len(self.children) > 0:
            child_max_paths = []
            for child in self.children:
                # print(["child"]*level)
                child_max_paths.append(child.compute_max_path_length(level + 1))
            return max(child_max_paths)
        return level

    def get_max_path_length(self):
        if self.max_path_length == 0:
            self.max_path_length = self.compute_max_path_length()
        return self.max_path_length

    def crop_orphans(self, max_orphan_count=4):
        favourite_children = []
        if len(self.children) > 0:
            counter = 0
            for child in self.children:
                if len(child.children) > 0 and counter < max_orphan_count:
                    favourite_children.append(child)
                    counter += 1
                    child.crop_orphans(max_orphan_count)
        self.children = favourite_children

    def all_tweet_ids(self):
        result = [self.post_id]
        for child in self.children:
            result.append(child.post_id)
        return result



class GraphNotInitializedException(Exception):
    def __init__(self, message="You need to call the initialize method before using metrics"):
        self.message = message
        super().__init__(self.message)


class NotATreeException(Exception):
    def __init__(self, message="NX Graph is Not A Tree"):
        self.message = message
        super().__init__(self.message)



import itertools

import networkx as nx

from delab_trees.exceptions import NotATreeException


def get_root(conversation_graph: nx.DiGraph):  # tree rooted at 0
    """
    :param conversation_graph:
    :return: the root node of a nx graph that is a tree
    """
    roots = [n for n, d in conversation_graph.in_degree() if d == 0]
    if len(roots) != 1:
        raise NotATreeException()
    return roots[0]


def get_all_reply_paths(conversation_graph: nx.DiGraph, min_path_length, required_max_path_length):
    """
    Get all reply paths that fall in the length window
    :param conversation_graph:
    :param min_path_length:
    :param required_max_path_length:
    :return:
    """
    G = conversation_graph
    all_paths = []
    nodes_combs = itertools.combinations(G.nodes, 2)
    for source, target in nodes_combs:
        paths = nx.all_simple_paths(G, source=source, target=target, cutoff=required_max_path_length)

        for path in paths:
            if path not in all_paths and path[::-1] not in all_paths and len(path) >= min_path_length:
                all_paths.append(path)
    return all_paths


def get_path(post_id, conversation_graph: nx.DiGraph, min_path_length=3, required_max_path_length=4):
    paths = get_all_reply_paths(conversation_graph, min_path_length, required_max_path_length)
    current_best_path_index = None
    current_best_score = 0
    index_count = 0
    for path in paths:
        if post_id in path:
            p_index = path.index(post_id)
            previous_tweets = p_index
            following_tweets = len(path) - p_index - 1
            middleness_score = min(previous_tweets, following_tweets) - abs(previous_tweets - following_tweets)
            if middleness_score > current_best_score:
                current_best_path_index = index_count
            current_best_score = max(current_best_score, middleness_score)
        index_count += 1
    if current_best_path_index is None:
        return None
    return paths[current_best_path_index]



import pickle
from random import choice

import pandas as pd

from delab_trees.algorithms.preperation_alg_pb import prepare_pb_data
from delab_trees.algorithms.training_alg_pb import train_pb
from delab_trees.constants import TREE_IDENTIFIER
from delab_trees.delab_tree import DelabTree
from delab_trees.algorithms.preperation_alg_rb import prepare_rb_data
from delab_trees.algorithms.training_alg_rb import train_rb


class TreeManager:

    def __init__(self, df):
        self.trees = {}
        self.df = df
        self.initialize_trees()

    def initialize_trees(self):
        grouped_by_tree_ids = {k: v for k, v in self.df.groupby(TREE_IDENTIFIER)}
        for tree_id, df in grouped_by_tree_ids.items():
            tree = DelabTree(df)
            tree.as_reply_graph()
            self.trees[tree_id] = tree
        return self

    def single(self) -> DelabTree:
        assert len(self.trees.keys()) == 1, "There needs to be exactly one tree in the manager!"
        return self.trees[self.trees.keys()[0]]

    def random(self) -> DelabTree:
        assert len(self.trees.keys()) >= 1
        return self.trees[choice(list(self.trees.keys()))]

    def __prepare_rb_model(self, prepared_data_filepath):
        """
        convert the trees to a matrix with all the node pairs (node -> ancestor) in order run the RB
        algorithm that wagers a prediction whether a post has been seen by a subsequent post's author

        """
        rb_prepared_data = prepare_rb_data(self)
        rb_prepared_data.to_pickle(prepared_data_filepath)
        return rb_prepared_data

    def __train_apply_rb_model(self, prepared_data_filepath="rb_prepared_data.pkl", prepare_data=True) \
            -> dict['tree_id', dict['author_id', 'rb_vision']]:
        if prepare_data:
            data = self.__prepare_rb_model(prepared_data_filepath)
        else:
            with open("rb_prepared_data.pkl", 'rb') as f:
                data = pickle.load(f)
        assert data.empty is False, "There was a mistake during the preparation of the data for the rb algorithm"
        applied_model, trained_model, features = train_rb(data)
        result = applied_model.pivot(columns='conversation_id', index='author', values="predictions").to_dict()
        return result

    def get_rb_vision(self, tree: DelabTree = None):
        applied_rb_model = self.__train_apply_rb_model()
        if tree is None:
            return applied_rb_model
        else:
            return applied_rb_model[tree.conversation_id]

    def __prepare_pb_model(self, prepared_data_filepath):
        rb_prepared_data = prepare_pb_data(self)
        rb_prepared_data.to_pickle(prepared_data_filepath)
        return rb_prepared_data

    def __train_apply_pb_model(self, prepared_data_filepath="pb_prepared_data.pkl", prepare_data=True) \
            -> dict['tree_id', dict['author_id', 'pb_vision']]:
        if prepare_data:
            data = self.__prepare_pb_model(prepared_data_filepath)
        else:
            with open("pb_prepared_data.pkl", 'rb') as f:
                data = pickle.load(f)
        assert data.empty is False, "There was a mistake during the preparation of the data for the pb algorithm"
        applied_model = train_pb(data)

        result = applied_model.pivot(columns='conversation_id', index='author', values="predictions").to_dict()
        return result

    def get_pb_vision(self, tree: DelabTree = None):
        applied_pb_model = self.__train_apply_pb_model()
        if tree is None:
            return applied_pb_model
        else:
            return applied_pb_model[tree.conversation_id]


def get_test_tree() -> DelabTree:
    from delab_trees import TreeManager

    d = {'tree_id': [1] * 4,
         'post_id': [1, 2, 3, 4],
         'parent_id': [None, 1, 2, 1],
         'author_id': ["james", "mark", "steven", "john"],
         'text': ["I am James", "I am Mark", " I am Steven", "I am John"],
         "created_at": [pd.Timestamp('2017-01-01T01'),
                        pd.Timestamp('2017-01-01T02'),
                        pd.Timestamp('2017-01-01T03'),
                        pd.Timestamp('2017-01-01T04')]}
    df = pd.DataFrame(data=d)
    manager = TreeManager(df)
    manager.initialize_trees()  # creates one tree
    test_tree = manager.random()
    return test_tree


def get_test_manager() -> TreeManager:
    d = {'tree_id': [1] * 4,
         'post_id': [1, 2, 3, 4],
         'parent_id': [None, 1, 2, 1],
         'author_id': ["james", "mark", "steven", "john"],
         'text': ["I am James", "I am Mark", " I am Steven", "I am John"],
         "created_at": [pd.Timestamp('2017-01-01T01'),
                        pd.Timestamp('2017-01-01T02'),
                        pd.Timestamp('2017-01-01T03'),
                        pd.Timestamp('2017-01-01T04')]}
    d2 = d.copy()
    d2["tree_id"] = [2] * 4
    d2['parent_id'] = [None, 1, 2, 3]
    d3 = d.copy()
    d3["tree_id"] = [3] * 4
    d3['parent_id'] = [None, 1, 1, 1]
    # a case where an author answers himself
    d4 = d.copy()
    d4["tree_id"] = [4] * 4
    d4["author_id"] = ["james", "james", "james", "john"]

    d5 = d.copy()
    d5["tree_id"] = [5] * 4
    d5['parent_id'] = [None, 1, 2, 3]
    d5["author_id"] = ["james", "james", "james", "john"]

    df1 = pd.DataFrame(data=d)
    df2 = pd.DataFrame(data=d2)
    df3 = pd.DataFrame(data=d3)
    df4 = pd.DataFrame(data=d4)
    df5 = pd.DataFrame(data=d5)
    df_list = [df1, df2, df3, df4, df5]
    df = pd.concat(df_list, ignore_index=True)
    manager = TreeManager(df)
    manager.initialize_trees()
    return manager


def get_social_media_trees(platform="twitter", context="test"):
    assert platform == "twitter" or platform == "reddit", "platform needs to be reddit or twitter!"
    if context == "test":
        file = "../delab_trees/data/dataset_reddit_no_text.pkl"
    else:
        file = "delab_trees/data/dataset_reddit_no_text.pkl"
    file = file.replace("reddit", platform)
    df = pd.read_pickle(file)
    manager = TreeManager(df)
    return manager.initialize_trees()



class AuthorMetric:
    def __init__(self, closeness_centrality, betweenness_centrality, katz_centrality, baseline_author_vision):
        self.closeness_centrality = closeness_centrality
        self.betweenness_centrality = betweenness_centrality
        self.katz_centrality = katz_centrality
        self.baseline_author_vision = baseline_author_vision


class AdvancedAuthorMetrics(AuthorMetric):
    def __init__(self, closeness_centrality, betweenness_centrality, katz_centrality, baseline_author_vision, rb_vision,
                 pb_vision):
        super(AuthorMetric, self).__init__(closeness_centrality,
                                           betweenness_centrality,
                                           katz_centrality,
                                           baseline_author_vision)
        self.rb_vision = rb_vision
        self.pb_vision = pb_vision



from collections import defaultdict
from copy import deepcopy

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
from networkx import MultiDiGraph
from networkx.drawing.nx_pydot import graphviz_layout
from pandas import DataFrame

from delab_trees.constants import TABLE, GRAPH
from delab_trees.delab_author_metric import AuthorMetric
from delab_trees.delab_post import DelabPosts, DelabPost
from delab_trees.exceptions import GraphNotInitializedException
from delab_trees.util import get_root


class DelabTree:

    def __init__(self, df: pd.DataFrame):
        self.df: DataFrame = deepcopy(df)
        self.reply_graph: MultiDiGraph = None
        self.author_graph: MultiDiGraph = None
        self.conversation_id = self.df.iloc[0][TABLE.COLUMNS.TREE_ID]
        # print("initialized")

    def branching_weight(self):
        if self.reply_graph is None:
            raise GraphNotInitializedException()
        return nx.tree.branching_weight(self.as_tree())

    def average_branching_factor(self):
        result = 2 * self.reply_graph.number_of_edges() / self.reply_graph.number_of_nodes()
        return result

    def total_number_of_posts(self):
        return len(self.df.index)

    def as_reply_graph(self):
        if self.reply_graph is None:
            node2creation = self.df.set_index(TABLE.COLUMNS.POST_ID).to_dict()[TABLE.COLUMNS.CREATED_AT]
            df = self.df[self.df[TABLE.COLUMNS.PARENT_ID].notna()]
            df = df.assign(label=GRAPH.LABELS.PARENT_OF)
            networkx_graph = nx.from_pandas_edgelist(df,
                                                     source=TABLE.COLUMNS.PARENT_ID,
                                                     target=TABLE.COLUMNS.POST_ID,
                                                     edge_attr='label',
                                                     create_using=nx.MultiDiGraph())
            nx.set_node_attributes(networkx_graph, GRAPH.SUBSETS.TWEETS, name="subset")  # rename to posts
            nx.set_node_attributes(networkx_graph, node2creation, name=TABLE.COLUMNS.CREATED_AT)
            self.reply_graph = networkx_graph
        return self.reply_graph

    def as_author_graph(self):
        """
        This computes the combined reply graph with the author_of relations included.
        :return:
        """
        if self.reply_graph is None:
            self.as_reply_graph()
        if self.author_graph is not None:
            return self.author_graph
        df = self.df.assign(label=GRAPH.LABELS.AUTHOR_OF)
        # print(df)
        networkx_graph = nx.from_pandas_edgelist(df, source="author_id", target="post_id", edge_attr='label',
                                                 create_using=nx.MultiDiGraph())
        author2authorlabel = dict([(author_id, GRAPH.SUBSETS.AUTHORS) for author_id in self.__get_author_ids()])
        nx.set_node_attributes(networkx_graph, author2authorlabel, name="subset")  # rename to posts
        # print(networkx_graph.edges(data=True))
        self.author_graph = nx.compose(self.reply_graph, networkx_graph)
        return self.author_graph

    def as_author_interaction_graph(self):
        """
        This computes the projected graph from the reply graph to the who answered whom graph (different nodes).
        This could be considered a unweighted bipartite projection.
        :return:
        """
        G = self.as_author_graph()
        # assuming the dataframe and the reply graph are two views on the same data!
        author_ids = self.__get_author_ids()

        G2 = nx.DiGraph()
        G2.add_nodes_from(author_ids)

        for a in author_ids:
            tw1_out_edges = G.out_edges(a, data=True)
            for _, tw1, out_attr in tw1_out_edges:
                tw2_out_edges = G.out_edges(tw1, data=True)
                for _, tw2, _ in tw2_out_edges:
                    in_edges = G.in_edges(tw2, data=True)
                    # since target already has a source, there can only be in-edges of type author_of
                    for reply_author, _, in_attr in in_edges:
                        if in_attr["label"] == GRAPH.LABELS.AUTHOR_OF:
                            assert reply_author in author_ids
                            if a != reply_author:
                                G2.add_edge(a, reply_author, label=GRAPH.LABELS.ANSWERED_BY)

        return G2

    def __get_author_ids(self):
        author_ids = set(self.df[TABLE.COLUMNS.AUTHOR_ID].tolist())
        return author_ids

    def as_tree(self):
        if self.reply_graph is None:
            raise GraphNotInitializedException()
        root = get_root(self.reply_graph)
        tree = nx.bfs_tree(self.reply_graph, root)
        return tree

    def as_post_list(self) -> list[DelabPost]:
        return DelabPosts.from_pandas(self.df)

    def as_recursive_tree(self):
        # TODO IMPLEMENT recursive_tree conversion
        # The recursive Tree has the tostring and toxml implemented
        pass

    def as_merged_self_answers_graph(self, return_deleted=False):
        posts_df = self.df[[TABLE.COLUMNS.POST_ID,
                            TABLE.COLUMNS.AUTHOR_ID,
                            TABLE.COLUMNS.CREATED_AT,
                            TABLE.COLUMNS.PARENT_ID]]
        posts_df = posts_df.sort_values(by=TABLE.COLUMNS.CREATED_AT)
        to_delete_list = []
        to_change_map = {}
        for row_index in posts_df.index.values:
            # we are not touching the root
            author_id, parent_author_id, parent_id, post_id = self.__get_table_row_as_names(posts_df, row_index)
            if parent_id is None or np.isnan(parent_id):
                continue
            # if a tweet is merged, ignore
            if post_id in to_delete_list:
                continue
            # if a tweet shares the author with its parent, deleted it
            if author_id == parent_author_id:
                to_delete_list.append(post_id)
            # if the parent has been deleted, find the next available parent
            else:
                current = row_index
                moving_post_id = deepcopy(post_id)
                moving_parent_id = deepcopy(parent_id)
                moving_parent_author_id = deepcopy(parent_author_id)
                while moving_parent_id in to_delete_list:
                    # we can make this assertion because we did not delete the root
                    moving_author_id, moving_parent_author_id, moving_parent_id, moving_post_id = \
                        self.__get_table_row_as_names(posts_df, current)
                    assert moving_parent_id is not None
                    current = posts_df.index[posts_df[TABLE.COLUMNS.POST_ID] == moving_parent_id].values[0]
                if moving_post_id != post_id:
                    to_change_map[post_id] = moving_parent_id

        # constructing the new graph
        G = nx.DiGraph()
        edges = []
        row_indexes2 = [row_index for row_index in posts_df.index
                        if posts_df.loc[row_index][TABLE.COLUMNS.POST_ID] not in to_delete_list]
        post_ids = list(posts_df.loc[row_indexes2][TABLE.COLUMNS.POST_ID])
        for row_index2 in row_indexes2:
            author_id, parent_author_id, parent_id, post_id = self.__get_table_row_as_names(posts_df, row_index2)
            if parent_id is not None and not np.isnan(parent_id):
                # if parent_id not in post_ids and parent_id not in to_delete_list:
                #     print("conversation {} has no root_node".format(self.conversation_id))
                if post_id in to_change_map:
                    new_parent = to_change_map[post_id]
                    if new_parent in post_ids:
                        edges.append((new_parent, post_id))
                    else:
                        G.remove_node(post_id)
                else:
                    edges.append((parent_id, post_id))

        assert len(edges) > 0, "there are no edges for conversation {}".format(self.conversation_id)
        G.add_edges_from(edges, label=GRAPH.LABELS.PARENT_OF)
        nx.set_node_attributes(G, GRAPH.SUBSETS.TWEETS, name="subset")
        # return G, to_delete_list, changed_nodes
        print("removed {} and changed {}".format(to_delete_list, to_change_map))
        if return_deleted:
            return G, to_delete_list, to_change_map
        return G

    def get_conversation_flows(self):
        """
        computes all flows (paths that lead from root to leaf) in the reply tree
        :rtype: object
        :return:
        """
        reply_tree = self.reply_graph
        root = get_root(reply_tree)
        leaves = [x for x in reply_tree.nodes() if reply_tree.out_degree(x) == 0]
        flows = []
        flow_dict = {}
        for leaf in leaves:
            paths = nx.all_simple_paths(reply_tree, root, leaf)
            flows.append(next(paths))
        for flow in flows:
            flow_name = str(flow[0]) + "_" + str(flow[-1])
            flow_tweets_frame = self.df[self.df[TABLE.COLUMNS.POST_ID].isin(flow)]
            flow_tweets = DelabPosts.from_pandas(flow_tweets_frame)
            flow_dict[flow_name] = flow_tweets

        name_of_longest = max(flow_dict, key=lambda x: len(set(flow_dict[x])))
        return flow_dict, name_of_longest

    def get_author_metrics(self):
        result = {}
        author_interaction_graph = self.as_author_interaction_graph()
        katz_centrality = nx.katz_centrality(author_interaction_graph)
        baseline_author_vision = self.get_baseline_author_vision()
        try:
            betweenness_centrality = nx.betweenness_centrality(author_interaction_graph)
        except ValueError:
            betweenness_centrality = {}

        author_ids = self.__get_author_ids()
        for author_id in author_ids:
            a_closeness_centrality = nx.closeness_centrality(author_interaction_graph, author_id)
            a_betweenness_centrality = betweenness_centrality.get(author_id, None)
            a_katz_centrality = katz_centrality.get(author_id, None)
            a_baseline_author_vision = baseline_author_vision.get(author_id, None)
            metric = AuthorMetric(a_closeness_centrality,
                                  a_betweenness_centrality,
                                  a_katz_centrality,
                                  a_baseline_author_vision)
            result[author_id] = metric

        return result

    def get_baseline_author_vision(self):
        author2baseline = {}
        author_interaction_graph, to_delete_list, to_change_map = self.as_merged_self_answers_graph(return_deleted=True)
        root = get_root(author_interaction_graph)
        post2author, author2posts = self.__get_author_post_map()
        for node in to_delete_list:
            post2author.pop(node, None)
        for author in author2posts.keys():
            n_posts = len(author2posts[author])
            root_distance_measure = 0
            reply_vision_measure = 0
            for post in author2posts[author]:
                if post in to_delete_list:
                    continue
                if post == root:
                    root_distance_measure += 1
                else:
                    path = next(nx.all_simple_paths(author_interaction_graph, root, post))
                    root_distance = len(path)
                    root_distance_measure += 0.25 ** root_distance
                    reply_vision_path_measure = 0

                    reply_paths = next(nx.all_simple_paths(author_interaction_graph, root, post))
                    for previous_tweet in reply_paths:
                        if previous_tweet != post:
                            path_to_previous = nx.all_simple_paths(author_interaction_graph, previous_tweet, post)
                            path_to_previous = next(path_to_previous)
                            path_length = len(path_to_previous)
                            reply_vision_path_measure += 0.5 ** path_length
                    reply_vision_path_measure = reply_vision_path_measure / len(reply_paths)
                    reply_vision_measure += reply_vision_path_measure
            root_distance_measure = root_distance_measure / n_posts
            reply_vision_measure = reply_vision_measure / n_posts
            # author2baseline[author] = (root_distance_measure + reply_vision_measure) / 2  # un-normalized
            author2baseline[author] = reply_vision_measure  # un-normalized
            baseline = author2baseline[author]
            assert 0 <= baseline <= 1
        return author2baseline

    def __get_author_post_map(self):
        tweet2author = my_dict = self.df.set_index(TABLE.COLUMNS.POST_ID)[TABLE.COLUMNS.AUTHOR_ID].to_dict()
        inverted_dict = defaultdict(list)
        for key, value in tweet2author.items():
            inverted_dict[value].append(key)
        author2posts = dict(inverted_dict)
        return tweet2author, author2posts

    def get_single_author_metrics(self, author_id):
        return self.get_author_metrics().get(author_id, None)

    @staticmethod
    def __get_table_row_as_names(posts_df, row_index):
        post_data = posts_df.loc[row_index]
        parent_id = post_data[TABLE.COLUMNS.PARENT_ID]
        post_id = post_data[TABLE.COLUMNS.POST_ID]
        author_id = post_data[TABLE.COLUMNS.AUTHOR_ID]
        parent_author_id = None
        # if parent_id is not None and np.isnan(parent_id) is False:
        parent_author_frame = posts_df[posts_df[TABLE.COLUMNS.POST_ID] == parent_id]
        if not parent_author_frame.empty:
            parent_author_id = parent_author_frame.iloc[0][TABLE.COLUMNS.AUTHOR_ID]
        return author_id, parent_author_id, parent_id, post_id

    def paint_reply_graph(self):
        tree = self.as_tree()
        pos = graphviz_layout(tree, prog="twopi")
        # add_attributes_to_plot(conversation_graph, pos, tree)
        nx.draw_networkx_labels(tree, pos)
        nx.draw(tree, pos)
        plt.show()
